import java.util.Random;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

/**
 * A simple class with a few methods to experiment with random number generation in Java.
 * @author Charles Cusack
 * @version 1 (2/22/2024)
 */
public class RandomTester
{

    /**
     * Generate a bunch of numbers in a given range and output the distribution of
     * how often each number was generated. The idea is to see whether or not the 
     * random number generator is doing a good job of being uniform.
     * @param range the upper limit of how large the numbers should be. Integers will be generated in
     *        the range [0,range), so not including range.
     * @param howMany the number of numbers to generate.
     */
    public void testRandomLimitedRange(int range, int howMany) {
        Random rand = new Random(); // The random number generator
        // A hash map to map numbers that are generated to how many times they are generated.
        HashMap<Integer,Integer> freq = new HashMap<Integer,Integer>();

        // Generate howMany random numbers and put them in the appropriate  place in the map.
        for(int i=0;i<howMany;i++) {
            // Generate the random number
            int num = rand.nextInt(range);
            // Get the number of times that number has already been generated from the map.
            Integer f = freq.get(num);

            if(f!=null) {
                // If f is not null, we put back into the map one more than the number of times
                // this number has been generated since we generated it again.
                freq.put(num,f+1);
            } else {
                // If f is null, that means the number was not in the map so has not been generated before.
                // So put it in the map as occuring 1 times since this is the first time.
                freq.put(num,1);
            }
        }

        // Print the results. 
        // We know that only values in [0,size) will be keys in the map, so we just ask for the value of each
        // of those keys. If the result is null, that keep is not in the map so the number of occurrences is 0.
        for(int i=0;i<range;i++) {
            // Get the number of times i was generated from the map.
            Integer numTimesGenerated = freq.get(i);
            // If it is null, it wasn't in the map so we print 0 occurrences.
            if(numTimesGenerated==null) {
                System.out.println(i+": 0");
            } else {
                System.out.println(i+": "+numTimesGenerated);
            }
        }
        System.out.println("----------------------");  // Just so we can mark each new run of the tests.

    }

    /**
     * Generate a bunch of numbers and output the distribution of
     * how often each number was generated if it was generated more than once. 
     * We do not print those that were generated once because that will be A LOT of numbers
     * and seeing how many were generated more than once gives us an idea of what is going on.
     * If not a lot of numbers print, then the generator is doing a good job of being uniform.
     * If some number show up with a large count, that would be bad.
     * @param howMany the number of numbers to generate.
     */
    public void testRandomFullRange(int howMany) {
        Random rand = new Random();
        HashMap<Integer,Integer> freq = new HashMap<Integer,Integer>();
        for(int i=0;i<howMany;i++) {
            // The only difference between the first method and this method is that we don't 
            // have a parameter to nextInt() so it will generate numbers in the full range of
            // ints in java, which includes both negative and positive integers.
            int num = rand.nextInt(); 
            // See documentation above--the same logic applies here.
            Integer f = freq.get(num);
            if(f!=null) {
                freq.put(num,f+1);
            } else {
                freq.put(num,1);
            }
        }

        // Here we do not knokw what the key values are--they could be any integer between -2^31 and 2^31-1.
        // Therefore we need to iterate over the key set.
        // Because this is a set coming from the HashSet, the numbers will not be in any obvious order.

        // Get the set of keys from the HasMap
        Set<Integer> keys = freq.keySet();

        // Iterate over the keys, printing if the frequency (the value in the HashMap for this key)
        // is more than 1.
        for(Integer key : keys) {    
            Integer numTimesGenerated = freq.get(key);
            // Only print if the frequency is at least 2.
            if(numTimesGenerated>1) {
                System.out.println(key+": "+numTimesGenerated);
            }
        }
        System.out.println("----------------------");  // Just so we can mark each new run of the tests.

    }

    /**
     * Identical to testRandom except for it uses a different method of iterating over the HashMap.
     * @param howMany the number of numbers to generate.
     */
    public void testRandomFullRangeV2(int howMany) {
        // The code is the same as above in the first half of this method.
        Random rand = new Random();
        HashMap<Integer,Integer> freq = new HashMap<Integer,Integer>();
        for(int i=0;i<howMany;i++) {
            int num = rand.nextInt(); 
            Integer f = freq.get(num);
            if(f!=null) {
                freq.put(num,f+1);
            } else {
                freq.put(num,1);
            }
        }

        // Get the set of entries from the HashMap.
        // Map.Entry is a class that stores (key,value) pairs.
        // See the Java API for more information.
        // I know the syntax here looks complicated, but it really ins't. Just walk through what it SHOULD
        // mean based on what you have already seen and it starts to make sense:
        // This is a Set whose entries are of type Map.Entry<Integer,Inetger>, similar
        // to Set<String> being a set of type String.
        // Map.Entry is just a simple class that stores (key,value) pairs, and the type of object 
        // that the key and value are need to be specified, just as they are in a HashMap.
        // In this case, the key and value are both Integers.
        // 
        Set<Map.Entry<Integer,Integer>> entries = freq.entrySet();

        // You should notice that for this application, this way of iterating is almost the same as
        // the previous method. The hard part is understanding the syntax and wrapping your head around 
        // what it means. Once you do, you will find this way to be more useful in some contexts than
        // iterating over the keySet.
        for(Map.Entry<Integer,Integer> entry : entries) {    
            // Only print if the frequency is at least 2.
            if(entry.getValue()>1) {
                System.out.println(entry.getKey()+": "+entry.getValue());
            }
        }
        System.out.println("----------------------");  // Just so we can mark each new run of the tests.

    }

}
